using System;
using System.IO;
using System.Globalization;
using System.Text.RegularExpressions;
using System.Resources;
using System.Diagnostics;
using System.Collections;
using System.Collections.Specialized;
using System.Reflection;

namespace GroupFinity.Greplace.FileInfoPlugin
{

	[LocalizedEvaluator("fileEvaluator", "file")]
	[LocalizedEvaluatorUsage("fileSize", "${file:size}" )]
	[LocalizedEvaluatorUsage("fileName", "${file:name}" )]
	[LocalizedEvaluatorUsage("filePath", "${file:path}" )]
	[LocalizedEvaluatorUsage("fileCreationTime", "${file:time-create}" )]
	[LocalizedEvaluatorUsage("fileLastWriteTime", "${file:time-write}" )]
	[LocalizedEvaluatorUsage("fileLastAccessTime", "${file:time-access}" )]
	[LocalizedEvaluatorUsage("fileCount", "${file:count(*.*)}" )]
	[LocalizedEvaluatorUsage("fileParentInformation", "${file:parent- }" )]
	[LocalizedEvaluatorUsage("fileParentPath", "${file:parent-path}" )]
	[LocalizedEvaluatorUsage("fileGrandparentInformation", "${file:parent-parent- }" )]
	[LocalizedEvaluatorUsage("fileVersion", "${file:version(fM.fm.fB.fP)}" )]
	public class FileInformationEvaluator : IEvaluator
	{
		public string Evaluate( IGreppedText text, Match match, string input)
		{
			try
			{
				int colon = input.IndexOf( ':' );
				string path;
				if ( colon<0 ) path = GetFullPathOf( text.Parent );
				else
				{
					path = input.Substring( colon+1 );
					input = input.Substring( 0, colon );
				}

				while( input.StartsWith( "parent-" ) )
				{
					path = Path.GetDirectoryName( path );
					input = input.Substring( "parent-".Length );
				}

				if ( input=="size" )
				{
					if ( !File.Exists( path ) ) return string.Empty;

					using( FileStream s = new FileStream( path, FileMode.Open, FileAccess.Read ) )
					{
						return s.Length.ToString();
					}
				}
				else if ( input=="name" ) return Path.GetFileName( path );
				else if ( input=="path" ) return path;
				else if ( input.StartsWith( "count" ) )
				{
					if ( !Directory.Exists( path ) ) return string.Empty;

					bool findFile = false, findDir = false;
					if ( input.StartsWith( "count-dir-file" ) )
					{
						input = input.Substring( "count-file-dir".Length );
						findFile = true; findDir = true;
					}
					else if ( input.StartsWith( "count-file-dir" ) )
					{
						input = input.Substring( "count-file-dir".Length );
						findFile = true; findDir = true;
					}
					else if ( input.StartsWith( "count-dir" ) )
					{
						input = input.Substring( "count-dir".Length );
						findFile = false; findDir = true;
					}
					else if ( input.StartsWith( "count-file" ) )
					{
						input = input.Substring( "count-file".Length );
						findFile = true; findDir = false;
					}
					else
					{
						input = input.Substring( "count".Length );
						findFile = true; findDir = true;
					}
					if ( input.Length==0 ) input = "*";
					else
					{
						input = input.Substring( 1, input.Length-2 );
					}

					string[] patterns = input.Split(';');

					int count = 0;
					foreach( string pattern in patterns )
					{
						if ( !findFile ) count += Directory.GetDirectories( path, pattern ).Length;
						else if ( !findDir ) count += Directory.GetFiles( path, pattern ).Length;
						else count += Directory.GetFileSystemEntries( path, pattern ).Length;
					}
					return count.ToString();
				}
				else if ( input.StartsWith( "time" ) )
				{
					FileSystemInfo info = File.Exists( path ) ? (FileSystemInfo)new FileInfo( path ):
						(FileSystemInfo)new DirectoryInfo( path );

					DateTime time = DateTime.Now;
					if ( input.StartsWith( "time-create" ) )
					{
						input = input.Substring( "time-create".Length );
						time = info.CreationTime;
					}
					else if ( input.StartsWith("time-access" ) )
					{
						input = input.Substring( "time-access".Length );
						time = info.LastAccessTime;
					}
					else
					{
						input = input.Substring( "time".Length );
						time = info.LastWriteTime;
					}
				
					if ( input.Length==0 ) input = "yyyy-MM-ddTHH:mm:sszzz";
					else
					{
						input = input.Substring( 1, input.Length-2 );
					}

					return time.ToString( input );
				}
				else if ( input.StartsWith( "version" ) )
				{
					input = input.Substring( "version".Length );

					if ( input.Length==0 ) input = "FM.Fm.Fb.Fp";
					else
					{
						input = input.Substring( 1, input.Length-2 );
					}

					FileVersionInfo info = FileVersionInfo.GetVersionInfo( path );
					string result = "";
					while( input.Length!=0 )
					{
						if ( input[0]=='"' )
						{
							int end = input.IndexOf( '"', 1 );
							if ( end<0 )
							{
								result += input;
								break;
							}
							else
							{
								result += input.Substring( 1, end-1 );
								input = input.Substring( end+1 );
								continue;
							}
						}

						if ( input[0]=='\'' )
						{
							int end = input.IndexOf( '\'', 1 );
							if ( end<0 )
							{
								result += input;
								break;
							}
							else
							{
								result += input.Substring( 1, end-1 );
								input = input.Substring( end+1 );
								continue;
							}
						}

						if ( input[0]=='\\' )
						{
							if ( input.Length==1 )
							{
								input = "";
								continue;
							}
							else
							{
								result += input[1];
								input = input.Substring( 2 );
								continue;
							}
						}

						bool found = false;
						foreach( string name in FileVersionMap.Keys )
						{
							if ( input.StartsWith( name ) )
							{
								input = input.Substring( name.Length );
								result += (FileVersionMap[name] as PropertyInfo).GetValue( info, new object[0] );
								found = true;
								break;
							}
						}
						if ( !found )
						{
							result += input[0];
							input = input.Substring( 1 );
						}
					}
					return result;
				}
				return string.Empty;
			}
			catch{ return string.Empty; }
		}
		
		private static readonly Hashtable FileVersionMap = new Hashtable();
		static FileInformationEvaluator()
		{
			FileVersionMap["fM"] = typeof( FileVersionInfo ).GetProperty( "FileMajorPart" );
			FileVersionMap["fm"] = typeof( FileVersionInfo ).GetProperty( "FileMinorPart" );
			FileVersionMap["fB"] = typeof( FileVersionInfo ).GetProperty( "FileBuildPart" );
			FileVersionMap["fP"] = typeof( FileVersionInfo ).GetProperty( "FilePrivatePart" );
			FileVersionMap["fD"] = typeof( FileVersionInfo ).GetProperty( "FileDescription" );
			FileVersionMap["fN"] = typeof( FileVersionInfo ).GetProperty( "FileName" );
			FileVersionMap["fV"] = typeof( FileVersionInfo ).GetProperty( "FileVersion" );
			FileVersionMap["pM"] = typeof( FileVersionInfo ).GetProperty( "ProductMajorPart" );
			FileVersionMap["pm"] = typeof( FileVersionInfo ).GetProperty( "ProductMinorPart" );
			FileVersionMap["pB"] = typeof( FileVersionInfo ).GetProperty( "ProductBuildPart" );
			FileVersionMap["pP"] = typeof( FileVersionInfo ).GetProperty( "ProductPrivatePart" );
			FileVersionMap["pN"] = typeof( FileVersionInfo ).GetProperty( "ProductName" );
			FileVersionMap["pV"] = typeof( FileVersionInfo ).GetProperty( "ProductVersion" );
			
			foreach( PropertyInfo info in typeof( FileVersionInfo ).GetProperties() )
			{
				FileVersionMap[info.Name] = info;
			}
		}

		private static string GetFullPathOf( IGreppedPath path )
		{
			return path.Parent==null ? path.Name : Path.Combine( GetFullPathOf( path.Parent ), path.Name );
		}
	}
}
