using System;
using System.Collections;
using System.Text.RegularExpressions;

namespace GroupFinity.Greplace.StandardPlugin
{
	[LocalizedEvaluator("calcEvaluator", "calc")]
	[LocalizedEvaluatorUsage("calc", "${calc: }" )]
	public class CalcEvaluator : IEvaluator
	{
		static readonly char[] Operators = new char[]
			{
				'`', '@',
				'*', '/', '%',
				'+', '-',
				'=', '!', '<', '>',
				'|', '&', 
		};
		static readonly int[] Priorities = new int[]
			{
				5, 5,
				4, 4, 4,
				3, 3,
				2, 2, 2, 2,
				1, 1,
		};

		static CalcEvaluator()
		{
			
		}

		public string Evaluate( IGreppedText text, Match match, string input)
		{
			try
			{
				return Calc( SplitString( input ) );
			}
			catch( Exception e )
			{
				System.Diagnostics.Debug.Assert( false, e.ToString() );
				return string.Empty;
			}
		}

		private static string Calc( ArrayList tokens )
		{
			SetPriorities( tokens );

			for(;;)
			{
				int ope = FindPrimaryOperator( tokens );
				if ( ope<0 ) break;
				decimal result = 0;
				decimal left = ope==0 ? 0 : ((Token)tokens[ope-1]).Value;
				decimal right = ope==tokens.Count-1 ? 0 : ((Token)tokens[ope+1]).Value;
				switch( Operators[(int)((Token)tokens[ope]).Value] )
				{
					case '`': result = (decimal)Math.Pow( (double)left, (double)right ); break;
					case '@': result = (decimal)Math.Log( (double)left, (double)right ); break;
					case '*': result = left*right; break;
					case '/': result = left/right; break;
					case '%': result = left%right; break;
					case '+': result = left+right; break;
					case '-': result = left-right; break;
					case '=': result = left==right ? 1 : 0; break;
					case '!': result = left!=right ? 1 : 0; break;
					case '<': result = left<right ? 1 : 0; break;
					case '>': result = left>right ? 1 : 0; break;
					case '|': result = left!=0 || right!=0 ? 1 : 0; break;
					case '&': result = left!=0 && right!=0 ? 1 : 0; break;
				}
				if ( ope!=tokens.Count-1 ) tokens.RemoveAt( ope+1 );
				tokens[ope] = new Token( result, TokenType.Number );
				if ( ope!=0 ) tokens.RemoveAt( ope-1 );
			}
			return ((Token)tokens[0]).Value.ToString();
		}

		class Token
		{
			public decimal Value;
			public TokenType Type;
			public int Priority;
			public Token( decimal value, TokenType type ){ Value = value; Type = type; Priority = 0; }
		}

		enum TokenType
		{
			Number,
			Operator,
			Bracket,
		}

		private static ArrayList SplitString( string input )
		{
			ArrayList result = new ArrayList();

			for( int i = 0; i<input.Length; )
			{
				if ( char.IsNumber( input[i] ) || input[i]=='.' || input[i]==',' )
				{
					int start = i;
					while( i<input.Length && 
						( char.IsNumber( input[i] ) || input[i]=='.' || input[i]==',' ) ) ++i;
					try
					{
						string number = input.Substring( start, i-start );
						decimal value = 0;
						if ( number.StartsWith( "0x" ) )
						{
							value = Convert.ToInt64( number.Substring( 2 ), 16 );
						}
						else if ( number.StartsWith( "0o" ) )
						{
							value = Convert.ToInt64( number.Substring( 2 ), 8 );
						}
						else if ( number.StartsWith( "0b" ) )
						{
							value = Convert.ToInt64( number.Substring( 2 ), 2 );
						}
						else
						{
							value = decimal.Parse( number, System.Globalization.NumberStyles.Any );
						}
						result.Add( new Token( value, TokenType.Number ) );
					}
					catch
					{
						result.Add( new Token( 0, TokenType.Number ) );
					}
				}
				else if ( input[i]=='(' || input[i]==')' )
				{
					result.Add( new Token( input[i], TokenType.Bracket ) );
					++i;
				}
				else
				{
					for( int j = 0; j<Operators.Length; ++j )
					{
						if ( Operators[j]==input[i] )
						{
							result.Add( new Token( j, TokenType.Operator ) );
						}
					}
					++i;
				}
			}

			return result;
		}

		private static void SetPriorities( ArrayList tokens )
		{
			int currentPriority = 0;
			for( int i = 0; i<tokens.Count; ++i )
			{
				Token token = (Token)tokens[i];
				if ( token.Type==TokenType.Bracket )
				{
					currentPriority += token.Value=='(' ? +10 : -10;
					tokens.RemoveAt( i );
					--i;
				}
				else if ( token.Type==TokenType.Operator ) token.Priority = currentPriority + Priorities[(int)token.Value];
			}
		}

		private static int FindPrimaryOperator( ArrayList tokens )
		{
			int maxIndex = -1;
			int maxPriority = 0;
			for( int i = 0; i<tokens.Count; ++i )
			{
				Token token = (Token)tokens[i];
				if ( token.Type==TokenType.Operator )
				{
					if ( token.Priority>maxPriority )
					{
						maxPriority = token.Priority;
						maxIndex = i;
					}
				}
			}
			return maxIndex;
		}
	}
}
