using System;
using System.Text;
using System.Text.RegularExpressions;
using System.Runtime.InteropServices;

namespace GroupFinity.Greplace.StandardPlugin
{
	[LocalizedEvaluator("stringEvaluator", "string")]
	[LocalizedEvaluatorUsage("stringLength", "${string:length: }" )]
	[LocalizedEvaluatorUsage("stringCaseLower", "${string:lower: }" )]
	[LocalizedEvaluatorUsage("stringCaseUpper", "${string:upper: }" )]
	[LocalizedEvaluatorUsage("stringWidthWiden", "${string:widen: }" )]
	[LocalizedEvaluatorUsage("stringWidthNarrow", "${string:narrow: }" )]
	[LocalizedEvaluatorUsage("stringJapaneseHiragana", "${string:hiragana: }" )]
	[LocalizedEvaluatorUsage("stringJapaneseKatakana", "${string:katakana: }" )]
	[LocalizedEvaluatorUsage("stringJapaneseHalfKatakana", "${string:katakana-narrow: }" )]
	[LocalizedEvaluatorUsage("stringChineseSimplify", "${string:simplify: }" )]
	[LocalizedEvaluatorUsage("stringChineseTraditional", "${string:traditional: }" )]
	public class StringEvaluator : IEvaluator
	{
		[DllImport("kernel32.dll")]
		static extern int LCMapString( int Locale, LCMAP dwMapFlags, string lpSrcStr, int cchSrc,
			StringBuilder lpDestStr, int cchDest );

		enum LCMAP : uint
		{
			LOWERCASE          = 0x00000100,  // lower case letters
			UPPERCASE          = 0x00000200,  // upper case letters
			SORTKEY            = 0x00000400,  // WC sort key (normalize)
			BYTEREV            = 0x00000800,  // byte reversal
			HIRAGANA           = 0x00100000,  // map katakana to hiragana
			KATAKANA           = 0x00200000,  // map hiragana to katakana
			HALFWIDTH          = 0x00400000,  // map double byte to single byte
			FULLWIDTH          = 0x00800000,  // map single byte to double byte
			LINGUISTIC_CASING  = 0x01000000,  // use linguistic rules for casing
			SIMPLIFIED_CHINESE = 0x02000000,  // map traditional chinese to simplified chinese
			TRADITIONAL_CHINESE = 0x04000000,  // map simplified chinese to traditional chinese
		}

		public string Evaluate( IGreppedText text, Match match, string input)
		{
			int begin = input.IndexOf( ':' );
			if ( begin==-1 ) return string.Empty;

			string[] operators = input.Substring( 0, begin ).Split( '-' );
			string result = input.Substring( begin+1 );

			int localeID = System.Globalization.CultureInfo.CurrentCulture.LCID;

			StringBuilder builder = new StringBuilder( result.Length*2+1 );

			for( int i = 0; i<operators.Length; ++i )
			{
				switch( operators[i] )
				{
					case "length": result = result.Length.ToString(); break;
					case "lower": result = result.ToLower(); break;
					case "upper": result = result.ToUpper(); break;
					case "widen":
						builder.EnsureCapacity( result.Length*2+1 );
						LCMapString( localeID, LCMAP.FULLWIDTH, result, -1, builder, builder.Capacity );
						result = builder.ToString().Replace("\\", "");
						break;
					case "narrow":
						builder.EnsureCapacity( result.Length*2+1 );
						LCMapString( localeID, LCMAP.HALFWIDTH, result, -1, builder, builder.Capacity );
						result = builder.ToString().Replace("", "\\");
						break;
					case "hiragana":
						builder.EnsureCapacity( result.Length*2+1 );
						LCMapString( localeID, LCMAP.HIRAGANA, result, -1, builder, builder.Capacity );
						result = builder.ToString();
						break;
					case "katakana":
						builder.EnsureCapacity( result.Length*2+1 );
						LCMapString( localeID, LCMAP.KATAKANA, result, -1, builder, builder.Capacity );
						result = builder.ToString();
						break;
					case "simplify":
						builder.EnsureCapacity( result.Length*2+1 );
						LCMapString( localeID, LCMAP.SIMPLIFIED_CHINESE, result, -1, builder, builder.Capacity );
						result = builder.ToString();
						break;
					case "traditional":
						builder.EnsureCapacity( result.Length*2+1 );
						LCMapString( localeID, LCMAP.TRADITIONAL_CHINESE, result, -1, builder, builder.Capacity );
						result = builder.ToString();
						break;
				}
			}

			return result;
		}
	}
}
